#include "Constant.h"
#include "math.h"
#include "MathStats.h"
#include <string>
#include <cstring>
#include <iostream>
#include "SingleVariantTest.h"
#include "Burden.h"
#include "snpHWE.h"
#define MATHLIB_STANDALONE 
#include <Rmath.h>

String SVT::geneMapfile = "refFlat_hg19.txt";
bool SVT::LRT = false;
bool SVT::SCORE = false;
bool SVT::WALD = false;
int SVT::Xstart = 2699520;
int SVT::Xend = 154931044;
bool SVT::dosage =false;
bool SVT::genoFromPed =false;
bool SVT::genoFromVCF =false;
bool SVT::FounderFreq =false;
String SVT::xLabel = "X";
int SVT::maleLabel = 1;
int SVT::femaleLabel =2;

bool SVT::GetGenotypeVectorPED(FastTransform & trans, Pedigree & ped,int markerid)
{
   bool status = false;
   mac =0.0;

   StringArray markerName;
   markerName.AddTokens(ped.markerNames[markerid],":");
   chr = markerName[0];
   pos = atoi(markerName[1].c_str());
   founderMaf = markerMaf = 0.0;

   int alleleCount = CountAlleles(ped, markerid);
   if(alleleCount!=2)
      return status;

   String ref = ped.GetMarkerInfo(markerid)->alleleLabels[1];
   String alt = ped.GetMarkerInfo(markerid)->alleleLabels[2];

   bool XStatus = false;
   if(AutoFit::fitX)
   {
      int found = markerName[0].FastFind(xLabel,0);
      if(found!=-1)
      {
	 if(pos > Xstart && pos < Xend)
	 {
	    XStatus=true;
	 }
      }
   }
   int n=0,nf=0;
   double fmaf = 0.0, maf=0.0;
   genotype.Dimension(trans.persons);
   for(int i=0;i<trans.persons;i++)
      genotype[i]=-1.0;

   //Loop through samples according to trans.pPheno 
   int idx=0;
   for(int f=0;f<ped.familyCount;f++)
   {
      for(int p=0;p<trans.pPheno[f].Length();p++)
      {
	 if(!ped[trans.pPheno[f][p]].isGenotyped(markerid))
	 {
	    idx++;
	    continue;
	 }
	 if(ped[trans.pPheno[f][p]].markers[markerid].isHeterozygous())
	 {
	    if(XStatus)
	    {
	       if(ped[trans.pPheno[f][p]].sex==SVT::femaleLabel)
	       {
		  genotype[idx] = 1.0;
	       }
	       if(ped[trans.pPheno[f][p]].sex==SVT::maleLabel)
	       {
		  genotype[idx] = 2.0;
	       }
	    }
	    else
	    {
	       genotype[idx] = 1.0;
	    }
	 }
	 else if(ped[trans.pPheno[f][p]].markers[markerid].isHomozygous())
	 {
	    if(ped[trans.pPheno[f][p]].markers[markerid].hasAllele(2))
	    {
	       genotype[idx] = 2.0;
	    }
	    else if(ped[trans.pPheno[f][p]].markers[markerid].hasAllele(1))
	    {
	       genotype[idx] = 0.0;
	    }
	 }
	 maf+=genotype[idx];
	 n++;
	 int founder = trans.foundersHash.Integer(ped[trans.pPheno[f][p]].pid);
	 if(founder != -1)
	 {
	    fmaf += genotype[idx];
	    nf++;
	 }
	 idx++;
      }
   }

   maf /= 2.0*n;
   fmaf /= 2.0*nf;
   if(fmaf==0.0 && maf>0)
      fmaf = 1.0/(2.0*nf);

   if(FounderFreq && (fmaf==0.0 || fmaf==1.0))
      return status;
   if(maf==0.0 || maf==1.0)
      return status;

   status=true;

   bool flip=false;
   if(FounderFreq && fmaf>0.5)
   {
      flip=true;
   }
   else if(maf>0.5) 
   {
      flip=true;
   }

   if(flip)
   {
      fmaf = 1.0-fmaf;
      maf = 1.0-maf;
      for(int i=0;i<trans.persons;i++)
      {
	 if(genotype[i]!=-1.0)
	 {
	    genotype[i] = 2.0-genotype[i];
	 }
      }
      rareAllele = ref;
      refAllele = alt;
   }
   else
   {
      rareAllele = alt;
      refAllele = ref;
   }

   founderMaf = fmaf;
   markerMaf = maf;

   for(int i=0;i<trans.persons;i++)
   {
      if(genotype[i]!=-1.0)
      {
	 mac+=genotype[i];
      }
   }

   for(int i=0;i<trans.persons;i++)
   {
      if(genotype[i]==-1.0)
      {
	 genotype[i] = 0.0;
      }
      else
      {
	 genotype[i] -= 2.0*maf;
      }
   }
   return status;
}

void SVT::Run(Pedigree & ped, FastTransform & trans,FastFit & engine, String & traitName,FILE * log)
{
   if(LRT || SCORE ||WALD) 
   {
      int traitNum = engine.GetTraitID(ped,traitName.c_str(),log);
      FILE * SCOREoutput;
      FILE * LRToutput;
      FILE * WALDoutput;
      String filename;
      if(SCORE)
      {
	 filename = Burden::outputFile + traitName + ".singlevar.score.txt";
	 SCOREoutput = fopen(filename,"wt");
	 PrintHeader(SCOREoutput,engine,trans);
      }
      if(LRT)
      {
	 filename = Burden::outputFile + traitName + ".singlevar.LRT.txt";
	 LRToutput = fopen(filename,"wt");
	 PrintHeader(LRToutput,engine,trans);
      }
      if(WALD)
      {
	 filename = Burden::outputFile + traitName + ".singlevar.wald.txt";
	 WALDoutput = fopen(filename,"wt");
	 PrintHeader(WALDoutput,engine,trans);
      }

      printf("\n\nNow doing SVT association tests for trait %s...\n",engine.traitName.c_str());
      Vector sigma2,transGeno; 
      transGeno.Dimension(trans.persons);
      sigma2.Dimension(trans.persons);
      double L0,L1,chisq,pvalue,effSize;

      int bad =0;
      if(LRT)
      {
	 if(AutoFit::fitSharedEnvironment || AutoFit::fitX)
	 {
	    L0=engine.EvaluateMulti(trans,false);
	 }
	 else
	 {
	    L0 = engine.Evaluate(engine.deltaHat,trans,false);
	 }
      }
      int markerCount =0;

      if(dosage || genoFromVCF)
      {
	 reader.open(GroupFromAnnotation::vcfInput,header);
	 int numSamples = header.getNumSamples();
	 while(reader.readRecord(record))
	 {
	    markerCount++;
	    if(!GetGenotypeVector(trans,ped,markerCount-1))
	    {
	       bad++;
	       if(LRT)
	       {
		  fprintf(LRToutput,"%s\t%d\t%g\t%g\t0\t-\t-\t-\t-\n",chr.c_str(),pos,founderMaf,markerMaf);
	       }
	       if(SCORE)
	       {
		  fprintf(SCOREoutput,"%s\t%d\t%g\t%g\t0\t-\t-\t-\t-\n",chr.c_str(),pos,founderMaf,markerMaf);
	       }
	       if(WALD)
	       {
		  fprintf(WALDoutput,"%s\t%d\t%g\t%g\t0\t-\t-\t-\t-\n",chr.c_str(),pos,founderMaf,markerMaf);
	       }
	       continue;
	    }
	    if(!trans.empKin)
	    {
	       transGeno.Zero();
	       int index=0;
	       for (int f = 0; f < ped.familyCount; f++)
		  if (trans.pPheno[f].Length() > 0)
		  {
		     int count = trans.pPheno[f].Length();
		     int block = index;
		     for(int i=0;i<count;i++)
		     {
			for(int c=0;c<count;c++)
			{
			   transGeno[index] += trans.transU[block+i][block+c]*genotype[block+c];
			}
			index++;
		     }
		  }
	    } else transGeno.Product(trans.transU,genotype);

	    if(LRT || WALD)
	    {
	       FastTransform  fullTrans;
	       FastFit fullEngine;

	       fullTrans.persons = trans.persons;
	       fullTrans.fixedEffects = trans.fixedEffects +1;
	       if(AutoFit::fitSharedEnvironment || AutoFit::fitX)
	       {
		  int covariateId = ped.GetCovariateID("MarkerTested");

		  fullTrans.pPheno = new IntArray [ped.familyCount];
		  // If we are using covariates, we only consider individuals
		  // for which all covariates have been recorded
		  for (int f = 0; f < ped.familyCount; f++)
		  {
		     fullTrans.pPheno[f].Dimension(0);
		     for (int i = ped.families[f]->first; i <= ped.families[f]->last; i++)
		     {
			if (ped[i].isPhenotyped(traitNum) && (!engine.useCovariates || ped[i].isFullyControlled()))
			{
			   fullTrans.pPheno[f].Push(i);
			}
		     }
		  }

		  int index =0;
		  for (int f = 0; f < ped.familyCount; f++)
		  {
		     if (fullTrans.pPheno[f].Length() > 0)
		     {
			int count = fullTrans.pPheno[f].Length();
			for(int i=0;i<count;i++)
			{
			   ped[fullTrans.pPheno[f][i]].covariates[ped.covariateCount-1] = genotype[index];
			   index++;
			}
		     }
		  }
		  AutoFit fullAutoEngine;
		  fullAutoEngine.useCovariates = true;
		  fullAutoEngine.fitSharedEnvironment = AutoFit::fitSharedEnvironment;
		  fullAutoEngine.fitX = AutoFit::fitX;

		  //Fit the alternative
		  fullAutoEngine.FitPolygenicModels(ped,traitNum,true);
		  effSize = fullAutoEngine.betaHat[fullAutoEngine.betaHat.Length()-1];

		  fullTrans.Y.Copy(trans.Y);
		  fullTrans.X.Copy(trans.X);
		  fullTrans.X.GrowTo(fullTrans.persons,trans.X.cols+1,0.0);
		  fullTrans.UY.Dimension(fullTrans.persons);
		  fullTrans.UX.Dimension(fullTrans.persons,fullTrans.fixedEffects);
		  fullEngine.beta.Dimension(fullTrans.fixedEffects);
		  fullEngine.betaHat.Dimension(fullTrans.fixedEffects);
		  fullTrans.transU.Dimension(0,0);
		  fullTrans.D.Dimension(0);

		  for(int i=0;i<fullTrans.persons;i++)
		  {
		     fullTrans.X[i][trans.fixedEffects] = genotype[i];
		  }
		  fullTrans.traitVar = fullAutoEngine.variance;
		  fullTrans.TransformPedkinX(fullAutoEngine, ped);

		  if(LRT)
		  {
		     fullEngine.betaHat.Copy(fullAutoEngine.betaHat);
		     L1=fullEngine.EvaluateMulti(fullTrans,false);

		     chisq = 2.0*(L1-L0);
		     //printf("L0 is:%g,L1 is:%g,chisq is: %g\n",L0,L1,chisq);
		     if(chisq<0)
			chisq = -chisq;
		     //printf("L0, L1, chisq = %g, %g, %g\n",L0,L1,chisq);
		     pvalue = pchisq(chisq, 1,0,0);
		     //pvalue = chidist(chisq, 1);
		     fprintf(LRToutput,"%s\t%d\t%g\t%g\t%g\t%s\t%s\t%7.3f\t%g\n",chr.c_str(),pos,founderMaf,markerMaf,mac,refAllele.c_str(),rareAllele.c_str(),effSize,pvalue);
		  }
		  if(WALD)
		  {
		     fullTrans.UDX.Dimension(fullTrans.persons,fullTrans.fixedEffects);
		     fullTrans.UDY.Dimension(fullTrans.persons);
		     fullTrans.inv.Dimension(fullTrans.fixedEffects,fullTrans.fixedEffects);

		     fullTrans.FinalizeTransform(0.0,0.0);
		     double chisq = effSize*effSize/fullTrans.inv[fullTrans.fixedEffects-1][fullTrans.fixedEffects-1];
		     pvalue = pchisq(chisq, 1,0,0);
		     //pvalue = chidist(chisq, 1);
		     fprintf(WALDoutput,"%s\t%d\t%g\t%g\t%g\t%s\t%s\t%7.3f\t%g\n",chr.c_str(),pos,founderMaf,markerMaf,mac,refAllele.c_str(),rareAllele.c_str(),effSize,pvalue);
		  }
	       }
	       else
	       {
		  fullTrans.D.Copy(trans.D);
		  fullTrans.UY.Copy(trans.UY);
		  fullTrans.transU.Copy(trans.transU);
		  fullTrans.UD.Dimension(fullTrans.persons,fullTrans.persons);

		  fullEngine.beta.Dimension(fullTrans.fixedEffects);
		  fullEngine.betaHat.Dimension(fullTrans.fixedEffects);
		  fullTrans.UX.Copy(trans.UX);
		  fullTrans.UX.GrowTo(trans.UX.rows,trans.UX.cols+1,0.0);
		  for(int row=0;row<fullTrans.persons;row++)
		  {
		     fullTrans.UX[row][fullTrans.fixedEffects-1] = transGeno[row];
		  }
		  fullEngine.OptimizeDelta(0.0001,fullTrans);
		  effSize = fullEngine.beta[fullTrans.fixedEffects-1];
		  if(LRT)
		  {
		     L1 = fullEngine.Evaluate(fullEngine.deltaHat,fullTrans,false);
		     chisq = 2.0*(L1-L0);
		     if(chisq<0)
			chisq = -chisq;
		     //printf("L0, L1, chisq = %g, %g, %g\n",L0,L1,chisq);
		     pvalue = pchisq(chisq, 1,0,0);
		     //pvalue = chidist(chisq, 1);
		     fprintf(LRToutput,"%s\t%d\t%g\t%g\t%g\t%s\t%s\t%7.3f\t%g\n",chr.c_str(),pos,founderMaf,markerMaf,mac,refAllele.c_str(),rareAllele.c_str(),effSize,pvalue);
		  }

		  if(WALD)
		  {
		     fullEngine.sigma_g2Hat = fullEngine.GetSigma_g(fullEngine.deltaHat,fullEngine.beta,fullTrans);
		     fullEngine.sigma_e2Hat = fullEngine.sigma_g2Hat * fullEngine.deltaHat;
		     //   printf("sigmag2 is %g;sigmae2 is %g\n",fullEngine.sigma_g2Hat,fullEngine.sigma_e2Hat);
		     fullTrans.UDX.Dimension(fullTrans.persons,fullTrans.fixedEffects);
		     fullTrans.UDY.Dimension(fullTrans.persons);
		     fullTrans.inv.Dimension(fullTrans.fixedEffects,fullTrans.fixedEffects);

		     fullTrans.FinalizeTransform(fullEngine.sigma_g2Hat,fullEngine.sigma_e2Hat);
		     double chisq = effSize*effSize/fullTrans.inv[fullTrans.fixedEffects-1][fullTrans.fixedEffects-1];
		     pvalue = pchisq(chisq, 1,0,0);
		     //pvalue = chidist(chisq, 1);
		     fprintf(WALDoutput,"%s\t%d\t%g\t%g\t%g\t%s\t%s\t%7.3f\t%g\n",chr.c_str(),pos,founderMaf,markerMaf,mac,refAllele.c_str(),rareAllele.c_str(),effSize,pvalue);
		  }
	       }
	    }
	    if(SCORE)
	    {
	       //           printf("genotype vector is:\n");
	       //           for(int i=0;i<genotype.Length();i++)
	       //              printf("%g ",genotype[i]);
	       //           printf("\n");
	       //printf("founderFreq is %d\n",FounderFreq);
	       double sigma_g2 = engine.sigma_g2Hat;
	       double sigma_e2 = engine.sigma_e2Hat;
	       if(AutoFit::fitSharedEnvironment || AutoFit::fitX)
	       {
		  for(int i=0;i<trans.persons;i++)
		     sigma2[i] = trans.D[i];
	       }
	       else
	       {
		  for(int i=0;i<trans.persons;i++)
		  {
		     sigma2[i] =sigma_g2*trans.D[i]+sigma_e2;
		  }
	       }
	       double numerator = 0.0;
	       double denominator = 1.0e-10;
	       double fixedEffect;
	       for(int i=0;i<trans.persons;i++)
	       {
		  fixedEffect=trans.UX[i].InnerProduct(engine.betaHat);
		  numerator += (trans.UY[i]-fixedEffect)*transGeno[i]/sigma2[i];
		  denominator += transGeno[i]*transGeno[i]/sigma2[i];
	       }

	       if(denominator==0)
	       {
		  fprintf(SCOREoutput,"%s\t%d\t%g\t%g\t%g\t%s\t%s\t-\t-\n",chr.c_str(),pos,founderMaf,markerMaf,mac,refAllele.c_str(),rareAllele.c_str()); 
		  continue;
	       }
	       else chisq = numerator*numerator/denominator;
	       //printf("numerator and denominator are: %g %g\n",numerator, denominator);
	       pvalue = pchisq(chisq, 1,0,0);
	       //pvalue = chidist(chisq, 1);
	       //This is the rare allele effect size
	       effSize = numerator/denominator;
	       fprintf(SCOREoutput,"%s\t%d\t%g\t%g\t%g\t%s\t%s\t%7.3f\t%g\n",chr.c_str(),pos,founderMaf,markerMaf,mac,refAllele.c_str(),rareAllele.c_str(),effSize,pvalue);
	    }
	 }
      }

      if(genoFromPed)
      {
	 markerCount = ped.markerCount;
	 for(int m = 0;m<ped.markerCount;m++) 
	 {
	    //markerName = ped.markerNames[m];
	    //printf("now testing marker %s: \n",markerName.c_str());
	    //Fill in grouped genotype vector here
	    //bool genostatus;
	    if(!GetGenotypeVector(trans,ped,m))  
	    {
	       bad++;
	       if(LRT)
	       {
		  fprintf(LRToutput,"%s\t%d\t%g\t%g\t0\t-\t-\t-\t-\n",chr.c_str(),pos,founderMaf,markerMaf);
	       }
	       if(SCORE)
	       {
		  fprintf(SCOREoutput,"%s\t%d\t%g\t%g\t0\t-\t-\t-\t-\n",chr.c_str(),pos,founderMaf,markerMaf);
	       }
	       if(WALD)
	       {
		  fprintf(WALDoutput,"%s\t%d\t%g\t%g\t0\t-\t-\t-\t-\n",chr.c_str(),pos,founderMaf,markerMaf);
	       }
	       continue;
	    }

	    if(!trans.empKin) 
	    {
	       transGeno.Zero();
	       int index=0; 
	       for (int f = 0; f < ped.familyCount; f++)
		  if (trans.pPheno[f].Length() > 0)
		  {
		     int count = trans.pPheno[f].Length();
		     int block = index;
		     for(int i=0;i<count;i++)
		     {
			for(int c=0;c<count;c++)
			{
			   transGeno[index] += trans.transU[block+i][block+c]*genotype[block+c];
			}
			index++;
		     }
		  }
	    } else transGeno.Product(trans.transU,genotype);

	    if(LRT || WALD)
	    {
	       FastTransform  fullTrans;
	       FastFit fullEngine;

	       fullTrans.persons = trans.persons;
	       fullTrans.fixedEffects = trans.fixedEffects +1;
	       if(AutoFit::fitSharedEnvironment || AutoFit::fitX)
	       {
		  int covariateId = ped.GetCovariateID("MarkerTested");

		  fullTrans.pPheno = new IntArray [ped.familyCount];
		  // If we are using covariates, we only consider individuals
		  // for which all covariates have been recorded
		  for (int f = 0; f < ped.familyCount; f++)
		  {
		     fullTrans.pPheno[f].Dimension(0);
		     for (int i = ped.families[f]->first; i <= ped.families[f]->last; i++)
		     {
			if (ped[i].isPhenotyped(traitNum) && (!engine.useCovariates || ped[i].isFullyControlled()))
			{
			   fullTrans.pPheno[f].Push(i);
			}
		     }
		  }

		  int index =0;
		  for (int f = 0; f < ped.familyCount; f++)
		  {
		     if (fullTrans.pPheno[f].Length() > 0)
		     {
			int count = fullTrans.pPheno[f].Length();
			for(int i=0;i<count;i++)
			{
			   ped[fullTrans.pPheno[f][i]].covariates[ped.covariateCount-1] = genotype[index];
			   index++;
			}
		     }
		  }

		  AutoFit fullAutoEngine; 
		  fullAutoEngine.useCovariates = true;
		  fullAutoEngine.fitSharedEnvironment = AutoFit::fitSharedEnvironment;
		  fullAutoEngine.fitX = AutoFit::fitX;

		  //Fit the alternative
		  fullAutoEngine.FitPolygenicModels(ped,traitNum,true);
		  effSize = fullAutoEngine.betaHat[fullAutoEngine.betaHat.Length()-1]; 

		  fullTrans.Y.Copy(trans.Y);
		  fullTrans.X.Copy(trans.X);
		  fullTrans.X.GrowTo(fullTrans.persons,trans.X.cols+1,0.0);
		  fullTrans.UY.Dimension(fullTrans.persons);
		  fullTrans.UX.Dimension(fullTrans.persons,fullTrans.fixedEffects);
		  fullEngine.beta.Dimension(fullTrans.fixedEffects);
		  fullEngine.betaHat.Dimension(fullTrans.fixedEffects);
		  fullTrans.transU.Dimension(0,0);
		  fullTrans.D.Dimension(0);

		  for(int i=0;i<fullTrans.persons;i++)
		  {
		     fullTrans.X[i][trans.fixedEffects] = genotype[i];
		  }
		  fullTrans.traitVar = fullAutoEngine.variance;
		  fullTrans.TransformPedkinX(fullAutoEngine, ped);

		  if(LRT)
		  {
		     fullEngine.betaHat.Copy(fullAutoEngine.betaHat);
		     L1=fullEngine.EvaluateMulti(fullTrans,false);

		     chisq = 2.0*(L1-L0);
		     //printf("L0 is:%g,L1 is:%g,chisq is: %g\n",L0,L1,chisq);
		     if(chisq<0)
			chisq = -chisq;
		     //printf("L0, L1, chisq = %g, %g, %g\n",L0,L1,chisq);
		     pvalue = pchisq(chisq, 1,0,0);
		     //pvalue = chidist(chisq, 1);
		     fprintf(LRToutput,"%s\t%d\t%g\t%g\t%g\t%s\t%s\t%7.3f\t%g\n",chr.c_str(),pos,founderMaf,markerMaf,mac,refAllele.c_str(),rareAllele.c_str(),effSize,pvalue);
		  }

		  if(WALD)
		  {
		     fullTrans.UDX.Dimension(fullTrans.persons,fullTrans.fixedEffects);
		     fullTrans.UDY.Dimension(fullTrans.persons);
		     fullTrans.inv.Dimension(fullTrans.fixedEffects,fullTrans.fixedEffects);

		     fullTrans.FinalizeTransform(0.0,0.0);
		     double chisq = effSize*effSize/fullTrans.inv[fullTrans.fixedEffects-1][fullTrans.fixedEffects-1];
		     pvalue = pchisq(chisq, 1,0,0);
		     //pvalue = chidist(chisq, 1);
		     fprintf(WALDoutput,"%s\t%d\t%g\t%g\t%g\t%s\t%s\t%7.3f\t%g\n",chr.c_str(),pos,founderMaf,markerMaf,mac,refAllele.c_str(),rareAllele.c_str(),effSize,pvalue);
		  }
	       }
	       else
	       {
		  fullTrans.D.Copy(trans.D);
		  fullTrans.UY.Copy(trans.UY);
		  fullTrans.transU.Copy(trans.transU);
		  fullTrans.UD.Dimension(fullTrans.persons,fullTrans.persons);

		  fullEngine.beta.Dimension(fullTrans.fixedEffects);
		  fullEngine.betaHat.Dimension(fullTrans.fixedEffects);
		  fullTrans.UX.Copy(trans.UX);
		  fullTrans.UX.GrowTo(trans.UX.rows,trans.UX.cols+1,0.0);
		  for(int row=0;row<fullTrans.persons;row++)
		  {
		     fullTrans.UX[row][fullTrans.fixedEffects-1] = transGeno[row];
		  }
		  fullEngine.OptimizeDelta(0.0001,fullTrans);
		  effSize = fullEngine.beta[fullTrans.fixedEffects-1];

		  if(LRT)
		  {
		     L1 = fullEngine.Evaluate(fullEngine.deltaHat,fullTrans,false);
		     chisq = 2.0*(L1-L0);
		     if(chisq<0)
			chisq = -chisq;
		     //printf("L0, L1, chisq = %g, %g, %g\n",L0,L1,chisq);
		     pvalue = pchisq(chisq, 1,0,0);
		     //pvalue = chidist(chisq, 1);
		     fprintf(LRToutput,"%s\t%d\t%g\t%g\t%g\t%s\t%s\t%7.3f\t%g\n",chr.c_str(),pos,founderMaf,markerMaf,mac,refAllele.c_str(),rareAllele.c_str(),effSize,pvalue);
		  }

		  if(WALD)
		  {
		     fullEngine.sigma_g2Hat = fullEngine.GetSigma_g(fullEngine.deltaHat,fullEngine.beta,fullTrans);
		     fullEngine.sigma_e2Hat = fullEngine.sigma_g2Hat * fullEngine.deltaHat;
		     //   printf("sigmag2 is %g;sigmae2 is %g\n",fullEngine.sigma_g2Hat,fullEngine.sigma_e2Hat);
		     fullTrans.UDX.Dimension(fullTrans.persons,fullTrans.fixedEffects);
		     fullTrans.UDY.Dimension(fullTrans.persons);
		     fullTrans.inv.Dimension(fullTrans.fixedEffects,fullTrans.fixedEffects);

		     fullTrans.FinalizeTransform(fullEngine.sigma_g2Hat,fullEngine.sigma_e2Hat);
		     double chisq = effSize*effSize/fullTrans.inv[fullTrans.fixedEffects-1][fullTrans.fixedEffects-1]; 
		     pvalue = pchisq(chisq, 1,0,0);
		     //pvalue = chidist(chisq, 1);
		     fprintf(WALDoutput,"%s\t%d\t%g\t%g\t%g\t%s\t%s\t%7.3f\t%g\n",chr.c_str(),pos,founderMaf,markerMaf,mac,refAllele.c_str(),rareAllele.c_str(),effSize,pvalue);
		  }
	       }
	    }

	    if(SCORE)
	    {
	       //	    printf("genotype vector is:\n");
	       //	    for(int i=0;i<genotype.Length();i++)
	       //	       printf("%g ",genotype[i]);
	       //	    printf("\n");
	       //printf("founderFreq is %d\n",FounderFreq);
	       double sigma_g2 = engine.sigma_g2Hat;
	       double sigma_e2 = engine.sigma_e2Hat;
	       if(AutoFit::fitSharedEnvironment || AutoFit::fitX)
	       {
		  for(int i=0;i<trans.persons;i++)
		     sigma2[i] = trans.D[i];
	       }
	       else
	       {
		  for(int i=0;i<trans.persons;i++)
		  {
		     sigma2[i] =sigma_g2*trans.D[i]+sigma_e2;
		  }
	       }
	       double numerator = 0.0;
	       double denominator = 1.0e-10; 
	       double fixedEffect;
	       for(int i=0;i<trans.persons;i++) 
	       {
		  fixedEffect=trans.UX[i].InnerProduct(engine.betaHat);
		  numerator += (trans.UY[i]-fixedEffect)*transGeno[i]/sigma2[i];
		  denominator += transGeno[i]*transGeno[i]/sigma2[i];
	       }

	       if(denominator==0)
	       {
		  fprintf(SCOREoutput,"%s\t%d\t%g\t%g\t%g\t%s\t%s\t-\t-\n",chr.c_str(),pos,founderMaf,markerMaf,mac,refAllele.c_str(),rareAllele.c_str());
		  continue;
	       }
	       else chisq = numerator*numerator/denominator;
	       //printf("numerator and denominator are: %g %g\n",numerator, denominator);
	       pvalue = pchisq(chisq, 1,0,0);
	       //pvalue = chidist(chisq, 1);
	       //This is the rare allele effect size
	       effSize = numerator/denominator;
	       fprintf(SCOREoutput,"%s\t%d\t%g\t%g\t%g\t%s\t%s\t%7.3f\t%g\n",chr.c_str(),pos,founderMaf,markerMaf,mac,refAllele.c_str(),rareAllele.c_str(),effSize,pvalue);
	    }
	 }
      }
      if(LRT)
      {
	 fprintf(LRToutput,"%d %s analyzed. %d of them %s not polymorphic among qualified individuals.\n",markerCount,markerCount>1?"markers were":"marker was",bad,bad>1?"are":"is");
      } 
      if(SCORE)
      {
	 fprintf(SCOREoutput,"%d %s analyzed. %d of them %s not polymorphic among qualified individuals.\n",markerCount,markerCount>1?"markers were":"marker was",bad,bad>1?"are":"is");
      }
      if(WALD)
      {
	 fprintf(WALDoutput,"%d %s analyzed. %d of them %s not polymorphic among qualified individuals.\n",markerCount,markerCount>1?"markers were":"marker was",bad,bad>1?"are":"is");
      }
      if(SCORE)
	 fclose(SCOREoutput);
      if(LRT)
	 fclose(LRToutput);
      if(WALD)
	 fclose(WALDoutput);

      //Restoring the pedigree covariate settings.
      if((AutoFit::fitX || AutoFit::fitSharedEnvironment) && (LRT || WALD))
      {
	 ped.covariateNames.Delete(ped.covariateNames.Length()-1);
	 ped.covariateLookup.Delete("MarkerTested");
	 ped.covariateCount--;
	 for(int i=0;i<ped.count;i++)
	    ped[i].covariates[ped.covariateCount] = 0.0;
      }
   }
}

void SVT::PrintHeader(FILE * output,FastFit & engine,FastTransform & trans)
{
   double sigma_g2 = engine.sigma_g2Hat;
   double h2 = 100.0*sigma_g2/trans.traitVar;
   double h2X = 100.0*engine.sigma_gXHat/trans.traitVar;
   double sharedEnv =100.0*engine.sigma_s2Hat/trans.traitVar;
   if(!AutoFit::fitSharedEnvironment && !AutoFit::fitX)
   {
      fprintf(output,"\n%s mean=%g, variance=%g, heritability=%g \n",engine.traitName.c_str(),trans.traitMean,trans.traitVar,h2);
   }
   else if(AutoFit::fitSharedEnvironment && AutoFit::fitX)
   {
      fprintf(output,"\n%s mean=%g, variance=%g, heritability=%g, Xheritability = %6.2f, SharedEnv = %6.2f \n",engine.traitName.c_str(),trans.traitMean,trans.traitVar,h2,h2X,sharedEnv);
   }
   else if(AutoFit::fitSharedEnvironment)
   {
      fprintf(output,"\n%s mean=%g, variance=%g, heritability=%g, SharedEnv = %g \n",engine.traitName.c_str(),trans.traitMean,trans.traitVar,h2,sharedEnv);
   }
   else if(AutoFit::fitX)
   {
      fprintf(output,"\n%s mean=%g, variance=%g, heritability=%g, Xheritability = %g \n",engine.traitName.c_str(),trans.traitMean,trans.traitVar,h2,h2X);
   }
   fprintf(output,"There are %d individuals included in the analysis.\n",trans.persons);
   fprintf(output,"===================================================================\n");
   fprintf(output,"\nCHR\tPOS\tFOUNDER_AF\tALL_AF\tINFORMATIVE_AC\tREF_ALLELE\tALT_ALLELE\tALT_ALLELE_EFFSIZE\tPVALUE\n");
}

bool SVT::GetGenotypeVectorVCFDosage(FastTransform & trans, Pedigree & ped)
{
   founderMaf = markerMaf = 0.0;
   bool status = false;
   String alt = record.getAltStr();
   String ref = record.getRefStr();
   chr = record.getChromStr();
   pos = record.get1BasedPosition();

   int n=0,nf=0;
   double fmaf = 0.0, maf=0.0;
   mac=0.0;

   VcfRecordGenotype & genoInfo = record.getGenotypeInfo();

   bool XStatus = false;
   if(AutoFit::fitX)
   {  
      if(chr==xLabel && pos > Xstart && pos < Xend)
      {  
	 XStatus=true;
      }  
   }

   genotype.Dimension(trans.persons);
   for(int i=0;i<trans.persons;i++)
      genotype[i]=-1.0;

   //Loop through samples according to trans.pPheno 
   //read in dosage into genotype vector
   int idx=0;
   for (int f = 0; f < ped.familyCount; f++)
   {
      for(int j=0;j<trans.pPheno[f].Length();j++)
      {
	 String sample = ped[trans.pPheno[f][j]].pid;

	 //get the sample number by sample ID from hash
	 int s = trans.sampleVCFIDHash.Integer(sample);
	 if(s!=-1)
	 {
	    const std::string * geno = genoInfo.getString("DS",s);
	    //if dosage is not missing
	    if(strcmp((*geno).c_str(),".")>0)
	    {
	       float dose = atof((*geno).c_str());
	       if(XStatus && dose >0 && ped[trans.pPheno[f][j]].sex==maleLabel)
	       {
		  genotype[idx] = 2;
	       }
	       else
		  genotype[idx] = dose;
	       maf += genotype[idx];
	       n++;

	       //judge if this is a founder
	       int founder = trans.foundersHash.Integer(sample);
	       if(founder != -1)
	       {
		  fmaf += genotype[idx];
		  nf++;
	       }
	    }
	 }
	 idx++;
      }
   }
   maf /= 2.0*n;
   fmaf /= 2.0*nf;
   if(fmaf==0.0 && maf>0.)
      fmaf = 1.0/(2.0*nf);

   if(FounderFreq && (fmaf==0.0 || fmaf==1.0))
      return status;
   if(maf==0.0 || maf==1.0)
      return status;

   status=true;

   bool flip=false;
   if(FounderFreq && fmaf>0.5)
   {
      flip=true;
   }
   else if(maf>0.5)
   {
      flip=true;
   }
   if(flip)
   {
      fmaf = 1.0-fmaf;
      maf = 1.0-maf;
      for(int i=0;i<trans.persons;i++)
      {
	 if(genotype[i]!=-1.0)
	 {
	    genotype[i] = 2.0-genotype[i];
	 }
      }
      rareAllele = ref;
      refAllele = alt;
   }
   else
   {
      rareAllele = alt;
      refAllele = ref;
   }
   founderMaf = fmaf;
   markerMaf = maf;

   for(int i=0;i<trans.persons;i++)
   {
      if(genotype[i]!=-1.0)
      {
	 mac+=genotype[i];
      }
   }

   for(int i=0;i<trans.persons;i++)
   {
      if(genotype[i]==-1.0)
      {
	 genotype[i] = 0.0;
      }
      else
      {
	 genotype[i] -= 2.0*maf;
      }
   }

   return status;
}

bool SVT::GetGenotypeVectorVCF(FastTransform & trans, Pedigree & ped)
{
   founderMaf = markerMaf = 0.0;
   bool status = false;

   String alt = record.getAltStr();
   String ref= record.getRefStr();
   chr = record.getChromStr();
   pos = record.get1BasedPosition();

   mac=0.0;
   double fmaf=0.0,maf=0.0;
   int n=0,nf=0;

   bool XStatus = false;
   if(AutoFit::fitX)
   {
      if(chr==xLabel && pos > Xstart && pos < Xend)
      {
	 XStatus=true;
      }
   }

   //Fill in genotype vector
   genotype.Dimension(trans.persons);
   for(int i=0;i<trans.persons;i++)
      genotype[i] = -1.0;

   int idx=0;
   for (int f = 0; f < ped.familyCount; f++)
   {
      for(int j=0;j<trans.pPheno[f].Length();j++)
      {
	 String sample = ped[trans.pPheno[f][j]].pid.c_str();

	 //get the sample number by sample ID from hash
	 int s = trans.sampleVCFIDHash.Integer(sample);
	 if(s!=-1)
	 {
	    int numGTs = record.getNumGTs(s);
	    Vector allele;

	    for(int jr= 0; jr < numGTs; jr++)
	    {
	       allele.Push(record.getGT(s,jr));
	    }
	    int bad =0;
	    for(int m=0;m<allele.Length();m++)
	    {
	       bad += (allele[m]==VcfGenotypeSample::MISSING_GT);
	    }
	    //if no missing allele exists
	    if(bad==0)
	    {
	       int founder = trans.foundersHash.Integer(sample);

	       int sum=0;
	       for(int i=0;i<allele.Length();i++)
		  sum+=allele[i];

	       if(XStatus && sum >0 && ped[trans.pPheno[f][j]].sex==maleLabel)
	       {
		  genotype[idx] = 2;
	       }
	       else
		  genotype[idx] = sum;
	       maf += genotype[idx];
	       n++;
	       if(founder!=-1)
	       {
		  nf++;
		  fmaf+=genotype[idx];
	       }
	    }
	 }
	 idx++;
      }
   }
   //Adjust genotype here
   maf /= 2.0*n;
   fmaf /=2.0*nf;

   if(fmaf==0.0 && maf>0.0)
      fmaf = 1.0/(2.0*nf);

   if(FounderFreq && (fmaf==0.0 || fmaf==1.0))
      return status;
   if(maf==0.0 || maf==1.0)
      return status;

   status=true;

   bool flip=false;
   if(FounderFreq && fmaf>0.5)
   {
      flip=true;
   }
   else if(maf>0.5)
   {
      flip=true;
   }

   if(flip)
   {
      fmaf = 1.0-fmaf;
      maf = 1.0-maf;
      for(int i=0;i<trans.persons;i++)
      {
	 if(genotype[i]!=-1.0)
	 {
	    genotype[i] = 2.0-genotype[i];
	 }
      }
      rareAllele = ref;
      refAllele = alt;
   }
   else
   {
      rareAllele = alt;
      refAllele = ref;
   }

   founderMaf = fmaf;
   markerMaf = maf;

   for(int i=0;i<trans.persons;i++)
   {
      if(genotype[i]!=-1.0)
      {
	 mac+=genotype[i];
      }
   }

   for(int i=0;i<trans.persons;i++)
   {
      if(genotype[i]==-1.0)
      {
	 genotype[i] = 0.0;
      }
      else
      {
	 genotype[i] -= 2.0*maf;
      }
   }
   return status;
}

bool SVT::GetGenotypeVector(FastTransform & trans, Pedigree & ped,int markerid)
{
   bool status;
   if(dosage)
   {
      status = GetGenotypeVectorVCFDosage(trans,ped);
   }
   if(genoFromPed)
   {
      status = GetGenotypeVectorPED(trans,ped,markerid);
   }
   if(genoFromVCF && !dosage)
   {
      status = GetGenotypeVectorVCF(trans,ped);
   }
   return status;
}


